use crate::host::{Host, ParseHostError};
use std::{fmt, str::FromStr};

#[derive(Clone, Debug, PartialEq)]
pub struct Request {
    user: String,
    host: Host,
    message: String,
}

impl fmt::Display for Request {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "misfin://{}@{} {}\r\n",
            self.user, self.host, self.message
        )
    }
}

#[derive(Clone, Debug, PartialEq)]
pub enum ParseRequestError {
    MissingSeparator,
    EmptyUser,
    EmptyHost,
    EmptyMessage,
    Malformed,
    ParseHostError(ParseHostError),
}

impl fmt::Display for ParseRequestError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::ParseHostError(e) => write!(f, "{e}"),
            _ => write!(f, "{self:?}"),
        }
    }
}

impl std::error::Error for ParseRequestError {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match self {
            Self::ParseHostError(e) => Some(e),
            _ => None,
        }
    }
}

impl From<ParseHostError> for ParseRequestError {
    fn from(value: ParseHostError) -> Self {
        Self::ParseHostError(value)
    }
}

impl FromStr for Request {
    type Err = ParseRequestError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if let Some((user, message)) = s.split_once(' ') {
            let Some(message) = message.strip_suffix("\r\n").map(ToString::to_string) else {
                return Err(ParseRequestError::Malformed);
            };
            if message.is_empty() {
                return Err(ParseRequestError::EmptyMessage);
            }
            if let Some((user, host)) = user.rsplit_once('@') {
                if host.is_empty() {
                    return Err(ParseRequestError::EmptyHost);
                } else if user == "misfin://" {
                    return Err(ParseRequestError::EmptyUser);
                }
                let host = host.parse()?;
                let Some(user) = user.strip_prefix("misfin://").map(ToString::to_string) else {
                    return Err(ParseRequestError::Malformed);
                };
                Ok(Request {
                    user,
                    host,
                    message,
                })
            } else {
                Err(ParseRequestError::MissingSeparator)
            }
        } else {
            Err(ParseRequestError::MissingSeparator)
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    const REQ_STR: &'static str = "misfin://john@misfin.example.com Anyone seen Jane?\r\n";

    fn req() -> Request {
        Request {
            user: "john".to_string(),
            host: Host {
                subdomain: Some("misfin".into()),
                domain: "example".into(),
                tld: "com".into(),
            },
            message: "Anyone seen Jane?".to_string(),
        }
    }

    #[test]
    fn to_string() {
        assert_eq!(req().to_string().as_str(), REQ_STR,);
    }

    #[test]
    fn parse() {
        assert_eq!(REQ_STR.parse::<Request>().unwrap(), req());
    }

    #[test]
    fn parse_missing_sep() {
        let req = "misfin://john@example.comHelloWorld!\r\n".parse::<Request>();
        assert_eq!(req, Err(ParseRequestError::MissingSeparator));
    }

    #[test]
    fn parse_malformed() {
        let req = "misfin://john@example.com Hello World!\n".parse::<Request>();
        assert_eq!(req, Err(ParseRequestError::Malformed));
        let req = "mail://john@example.com Hello World!\r\n".parse::<Request>();
        assert_eq!(req, Err(ParseRequestError::Malformed));
    }

    #[test]
    fn parse_empty_user() {
        let req = "misfin://@example.com Hello World!\r\n".parse::<Request>();
        assert_eq!(req, Err(ParseRequestError::EmptyUser));
    }

    #[test]
    fn parse_empty_host() {
        let req = "misfin://john@ Hello World!\r\n".parse::<Request>();
        assert_eq!(req, Err(ParseRequestError::EmptyHost));
    }

    #[test]
    fn parse_empty_msg() {
        let req = "misfin://john@example.com \r\n".parse::<Request>();
        assert_eq!(req, Err(ParseRequestError::EmptyMessage));
    }

    #[test]
    fn parse_host_illegal_whitespace() {
        let req = "misfin://john@example\tfairy.com Hello World!\r\n".parse::<Request>();
        assert_eq!(
            req,
            Err(ParseRequestError::ParseHostError(
                ParseHostError::IllegalWhitespace
            ))
        );
    }
}
